@@ -25,7 +25,18 @@ import sys
 
 import ecdsa
 import esptool
-import pyaes
+
+try:  # use pycrypto API if available
+    from Crypto.Cipher import AES
+
+    def ECB(key):
+        return AES.new(key, AES.MODE_ECB)
+
+except ImportError:
+    import pyaes
+
+    def ECB(key):
+        return pyaes.AESModeOfOperationECB(key)
 
 
 def get_chunks(source, chunk_len):
@@ -47,6 +58,25 @@ def swap_word_order(source):
     return struct.pack(words, *reversed(struct.unpack(words, source)))
 
 
+def _load_hardware_key(keyfile):
+    """ Load a 256-bit key, similar to stored in efuse, from a file
+
+    192-bit keys will be extended to 256-bit using the same algorithm used
+    by hardware if 3/4 Coding Scheme is set.
+    """
+    key = keyfile.read()
+    if len(key) not in [24, 32]:
+        raise esptool.FatalError("Key file contains wrong length (%d bytes), 24 or 32 expected." % len(key))
+    if len(key) == 24:
+        key = key + key[8:16]
+        print("Using 192-bit key (extended)")
+    else:
+        print("Using 256-bit key")
+
+    assert len(key) == 32
+    return key
+
+
 def digest_secure_bootloader(args):
     """ Calculate the digest of a bootloader image, in the same way the hardware
     secure boot engine would do so. Can be used with a pre-loaded key to update a
@@ -72,7 +102,7 @@ def digest_secure_bootloader(args):
     # if image isn't 128 byte multiple then pad with 0xFF (ie unwritten flash)
     # as this is what the secure boot engine will see
     if len(plaintext_image) % 128 != 0:
-        plaintext_image += "\xFF" * (128 - (len(plaintext_image) % 128))
+        plaintext_image += b"\xFF" * (128 - (len(plaintext_image) % 128))
 
     plaintext = iv + plaintext_image
 
@@ -81,10 +111,8 @@ def digest_secure_bootloader(args):
     # produce the digest. Each block in/out of ECB is reordered
     # (due to hardware quirks not for security.)
 
-    key = args.keyfile.read()
-    if len(key) != 32:
-        raise esptool.FatalError("Key file contains wrong length (%d bytes), 32 expected." % len(key))
-    aes = pyaes.AESModeOfOperationECB(key)
+    key = _load_hardware_key(args.keyfile)
+    aes = ECB(key)
     digest = hashlib.sha512()
 
     for block in get_chunks(plaintext, 16):
@@ -119,7 +147,7 @@ def generate_signing_key(args):
     print("ECDSA NIST256p private key in PEM format written to %s" % args.keyfile)
 
 
-def _load_key(args):
+def _load_ecdsa_signing_key(args):
     sk = ecdsa.SigningKey.from_pem(args.keyfile.read())
     if sk.curve != ecdsa.NIST256p:
         raise esptool.FatalError("Signing key uses incorrect curve. ESP32 Secure Boot only supports NIST256p (openssl calls this curve 'prime256v1")
@@ -128,7 +156,7 @@ def _load_key(args):
 
 def sign_data(args):
     """ Sign a data file with a ECDSA private key, append binary signature to file contents """
-    sk = _load_key(args)
+    sk = _load_ecdsa_signing_key(args)
 
     # calculate signature of binary data
     binary_content = args.datafile.read()
@@ -152,12 +180,18 @@ def sign_data(args):
 
 def verify_signature(args):
     """ Verify a previously signed binary image, using the ECDSA public key """
-    try:
-        sk = _load_key(args)  # try to load as private key first
+    key_data = args.keyfile.read()
+    if b"-BEGIN EC PRIVATE KEY" in key_data:
+        sk = ecdsa.SigningKey.from_pem(key_data)
         vk = sk.get_verifying_key()
-    except Exception:  # this is a catchall because ecdsa can throw private Exceptions
-        args.keyfile.seek(0)
-        vk = ecdsa.VerifyingKey.from_pem(args.keyfile.read())
+    elif b"-BEGIN PUBLIC KEY" in key_data:
+        vk = ecdsa.VerifyingKey.from_pem(key_data)
+    elif len(key_data) == 64:
+        vk = ecdsa.VerifyingKey.from_string(key_data,
+                                            curve=ecdsa.NIST256p)
+    else:
+        raise esptool.FatalError("Verification key does not appear to be an EC key in PEM format or binary EC public key data. Unsupported")
+
     if vk.curve != ecdsa.NIST256p:
         raise esptool.FatalError("Public key uses incorrect curve. ESP32 Secure Boot only supports NIST256p (openssl calls this curve 'prime256v1")
 
@@ -178,19 +212,25 @@ def verify_signature(args):
 
 def extract_public_key(args):
     """ Load an ECDSA private key and extract the embedded public key as raw binary data. """
-    sk = _load_key(args)
+    sk = _load_ecdsa_signing_key(args)
     vk = sk.get_verifying_key()
     args.public_keyfile.write(vk.to_string())
     print("%s public key extracted to %s" % (args.keyfile.name, args.public_keyfile.name))
 
 
 def digest_private_key(args):
-    sk = _load_key(args)
+    sk = _load_ecdsa_signing_key(args)
     repr(sk.to_string())
     digest = hashlib.sha256()
     digest.update(sk.to_string())
-    args.digest_file.write(digest.digest())
-    print("SHA-256 digest of private key %s written to %s" % (args.keyfile.name, args.digest_file.name))
+    result = digest.digest()
+    if args.keylen == 192:
+        result = result[0:24]
+    args.digest_file.write(result)
+    print("SHA-256 digest of private key %s%s written to %s" % (args.keyfile.name,
+                                                                "" if args.keylen == 256
+                                                                else " (truncated to 192 bits)",
+                                                                args.digest_file.name))
 
 
 # flash encryption key tweaking pattern: the nth bit of the key is
@@ -233,6 +273,30 @@ def _flash_encryption_tweak_range(flash_crypt_config=0xF):
     return tweak_range
 
 
+def _flash_encryption_tweak_range_bits(flash_crypt_config=0xF):
+    """ Return bits (in reverse order) that the "key tweak" applies to,
+    as determined by the FLASH_CRYPT_CONFIG 4 bit efuse value.
+    """
+    tweak_range = 0
+    if (flash_crypt_config & 1) != 0:
+        tweak_range |= 0xFFFFFFFFFFFFFFFFE00000000000000000000000000000000000000000000000
+    if (flash_crypt_config & 2) != 0:
+        tweak_range |= 0x00000000000000001FFFFFFFFFFFFFFFF0000000000000000000000000000000
+    if (flash_crypt_config & 4) != 0:
+        tweak_range |= 0x000000000000000000000000000000000FFFFFFFFFFFFFFFE000000000000000
+    if (flash_crypt_config & 8) != 0:
+        tweak_range |= 0x0000000000000000000000000000000000000000000000001FFFFFFFFFFFFFFF
+    return tweak_range
+
+
+# Forward bit order masks
+mul1        = 0x0000200004000080000004000080001000000200004000080000040000800010
+mul2        = 0x0000000000000000200000000000000010000000000000002000000000000001
+
+mul1_mask   = 0xffffffffffffff801ffffffffffffff00ffffffffffffff81ffffffffffffff0
+mul2_mask   = 0x000000000000007fe00000000000000ff000000000000007e00000000000000f
+
+
 def _flash_encryption_tweak_key(key, offset, tweak_range):
     """Apply XOR "tweak" values to the key, derived from flash offset
     'offset'. This matches the ESP32 hardware flash encryption.
@@ -243,37 +307,50 @@ def _flash_encryption_tweak_key(key, offset, tweak_range):
 
     Return tweaked key
     """
-    key = [ord(k) for k in key]
-    offset_bits = [(offset & (1 << x)) != 0 for x in range(24)]
+    if esptool.PYTHON2:
+        key = [ord(k) for k in key]
+        assert len(key) == 32
 
-    for bit in tweak_range:
-        if offset_bits[_FLASH_ENCRYPTION_TWEAK_PATTERN[bit]]:
-            # note that each byte has a backwards bit order, compared
-            # to how it is looked up in the tweak pattern table
-            key[bit // 8] ^= 1 << (7 - (bit % 8))
+        offset_bits = [(offset & (1 << x)) != 0 for x in range(24)]
 
-    return b"".join(chr(k) for k in key)
+        for bit in tweak_range:
+            if offset_bits[_FLASH_ENCRYPTION_TWEAK_PATTERN[bit]]:
+                # note that each byte has a backwards bit order, compared
+                # to how it is looked up in the tweak pattern table
+                key[bit // 8] ^= 1 << (7 - (bit % 8))
+
+        key = b"".join(chr(k) for k in key)
+        return key
+
+    else:
+        addr = offset >> 5
+        key ^= ((mul1 * addr) | ((mul2 * addr) & mul2_mask)) & tweak_range
+        return int.to_bytes(key, length=32, byteorder='big', signed=False)
 
 
 def generate_flash_encryption_key(args):
-    args.key_file.write(os.urandom(32))
+    print("Writing %d random bits to key file %s" % (args.keylen, args.key_file.name))
+    args.key_file.write(os.urandom(args.keylen // 8))
 
 
 def _flash_encryption_operation(output_file, input_file, flash_address, keyfile, flash_crypt_conf, do_decrypt):
-    key = keyfile.read()
-    if len(key) != 32:
-        raise esptool.FatalError("Key file contains wrong length (%d bytes), 32 expected." % len(key))
+    key = _load_hardware_key(keyfile)
 
     if flash_address % 16 != 0:
         raise esptool.FatalError("Starting flash address 0x%x must be a multiple of 16" % flash_address)
 
     if flash_crypt_conf == 0:
         print("WARNING: Setting FLASH_CRYPT_CONF to zero is not recommended")
-    tweak_range = _flash_encryption_tweak_range(flash_crypt_conf)
+
+    if esptool.PYTHON2:
+        tweak_range = _flash_encryption_tweak_range(flash_crypt_conf)
+    else:
+        tweak_range = _flash_encryption_tweak_range_bits(flash_crypt_conf)
+        key = int.from_bytes(key, byteorder='big', signed=False)
 
     aes = None
+    block_offs = flash_address
     while True:
-        block_offs = flash_address + input_file.tell()
         block = input_file.read(16)
         if len(block) == 0:
             break
@@ -282,12 +359,12 @@ def _flash_encryption_operation(output_file, input_file, flash_address, keyfile,
                 raise esptool.FatalError("Data length is not a multiple of 16 bytes")
             pad = 16 - len(block)
             block = block + os.urandom(pad)
-            print("WARNING: Padding with %d bytes of random data (encrypted data must be multiple of 16 bytes long)" % pad)
+            print("Note: Padding with %d bytes of random data (encrypted data must be multiple of 16 bytes long)" % pad)
 
         if (block_offs % 32 == 0) or aes is None:
             # each bit of the flash encryption key is XORed with tweak bits derived from the offset of 32 byte block of flash
             block_key = _flash_encryption_tweak_key(key, block_offs, tweak_range)
-            aes = pyaes.AESModeOfOperationECB(block_key)
+            aes = ECB(block_key)
 
         block = block[::-1]  # reverse input block byte order
 
@@ -301,6 +378,7 @@ def _flash_encryption_operation(output_file, input_file, flash_address, keyfile,
 
         block = block[::-1]  # reverse output block byte order
         output_file.write(block)
+        block_offs += len(block)
 
 
 def decrypt_flash_data(args):
@@ -341,24 +419,29 @@ def main():
 
     p = subparsers.add_parser('verify_signature',
                               help='Verify a data file previously signed by "sign_data", using the public key.')
-    p.add_argument('--keyfile', '-k', help="Public key file for verification. Can be the private key file (public key is embedded).",
+    p.add_argument('--keyfile', '-k', help="Public key file for verification. Can be private or public key in PEM format, " +
+                   "or a binary public key produced by extract_public_key command.",
                    type=argparse.FileType('rb'), required=True)
     p.add_argument('datafile', help="Signed data file to verify signature.", type=argparse.FileType('rb'))
 
     p = subparsers.add_parser('extract_public_key',
                               help='Extract the public verification key for signatures, save it as a raw binary file.')
-    p.add_argument('--keyfile', '-k', help="Private key file to extract the public verification key from.", type=argparse.FileType('rb'),
+    p.add_argument('--keyfile', '-k', help="Private key file (PEM format) to extract the public verification key from.", type=argparse.FileType('rb'),
                    required=True)
-    p.add_argument('public_keyfile', help="File to save new public key) into", type=argparse.FileType('wb'))
+    p.add_argument('public_keyfile', help="File to save new public key into", type=argparse.FileType('wb'))
 
     p = subparsers.add_parser('digest_private_key', help='Generate an SHA-256 digest of the private signing key. ' +
                               'This can be used as a reproducible secure bootloader or flash encryption key.')
-    p.add_argument('--keyfile', '-k', help="Private key file to generate a digest from.", type=argparse.FileType('rb'),
+    p.add_argument('--keyfile', '-k', help="Private key file (PEM format) to generate a digest from.", type=argparse.FileType('rb'),
                    required=True)
+    p.add_argument('--keylen', '-l', help="Length of private key digest file to generate (in bits). 3/4 Coding Scheme requires 192 bit key.",
+                   choices=[192, 256], default=256, type=int)
     p.add_argument('digest_file', help="File to write 32 byte digest into", type=argparse.FileType('wb'))
 
     p = subparsers.add_parser('generate_flash_encryption_key', help='Generate a development-use 32 byte flash encryption key with random data.')
-    p.add_argument('key_file', help="File to write 32 byte digest into", type=argparse.FileType('wb'))
+    p.add_argument('--keylen', '-l', help="Length of private key digest file to generate (in bits). 3/4 Coding Scheme requires 192 bit key.",
+                   choices=[192, 256], default=256, type=int)
+    p.add_argument('key_file', help="File to write 24 or 32 byte digest into", type=argparse.FileType('wb'))
 
     p = subparsers.add_parser('decrypt_flash_data', help='Decrypt some data read from encrypted flash (using known key)')
     p.add_argument('encrypted_file', help="File with encrypted flash contents", type=argparse.FileType('rb'))
