@@ -1,7 +1,9 @@
 /*
- * Copyright (c) 2016 Cesanta Software Limited & Espressif Systems (Shanghai) PTE LTD
+ * Copyright (c) 2016-2019 Espressif Systems (Shanghai) PTE LTD & Cesanta Software Limited
  * All rights reserved
  *
+ * This file is part of the esptool.py binary flasher stub.
+ *
  * This program is free software; you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free Software
  * Foundation; either version 2 of the License, or (at your option) any later version.
@@ -16,21 +18,15 @@
  */
 
 /*
- * Spiffy flasher. Implements strong checksums (MD5) and can use higher
- * baud rates. Actual max baud rate will differ from device to device,
- * but 921K seems to be common.
+ * Main flasher stub logic
  *
- * SLIP protocol is used for communication.
- * First packet is a single byte - command number.
- * After that, a packet with a variable number of 32-bit (LE) arguments,
- * depending on command.
+ * This stub uses the same SLIP framing and basic command/response structure
+ * as the in-ROM flasher program, but with some enhanced
+ * functions and also standardizes the flasher features between different chips.
  *
- * Then command produces variable number of packets of output, but first
- * packet of length 1 is the response code: 0 for success, non-zero - error.
- *
- * See individual command description below.
+ * Actual command handlers are implemented in stub_commands.c
  */
-
+#include <stdlib.h>
 #include "stub_flasher.h"
 #include "rom_functions.h"
 #include "slip.h"
@@ -40,16 +36,6 @@
 
 #define UART_RX_INTS (UART_RXFIFO_FULL_INT_ENA | UART_RXFIFO_TOUT_INT_ENA)
 
-
-#ifdef ESP32
-/* ESP32 register naming is a bit more consistent */
-#define UART_INT_CLR(X) UART_INT_CLR_REG(X)
-#define UART_INT_ST(X) UART_INT_ST_REG(X)
-#define UART_INT_ENA(X) UART_INT_ENA_REG(X)
-#define UART_STATUS(X) UART_STATUS_REG(X)
-#define UART_FIFO(X) UART_FIFO_REG(X)
-#endif
-
 static uint32_t get_new_uart_divider(uint32_t current_baud, uint32_t new_baud)
 {
   uint32_t master_freq;
@@ -58,7 +44,7 @@ static uint32_t get_new_uart_divider(uint32_t current_baud, uint32_t new_baud)
      is correct wrt the relative crystal accuracy of the ESP & the USB/serial adapter).
      From this we can estimate crystal freq, and update for a new baud rate relative to that.
   */
-  uint32_t uart_reg = REG_READ(UART_CLKDIV_REG(0));
+  uint32_t uart_reg = READ_REG(UART_CLKDIV_REG(0));
   uint32_t uart_div = uart_reg & UART_CLKDIV_M;
 #ifdef ESP32
   // account for fractional part of divider (bottom 4 bits)
@@ -119,18 +105,18 @@ static void uart_isr_receive(char byte)
 }
 
 void uart_isr(void *arg) {
-  uint32_t int_st = READ_PERI_REG(UART_INT_ST(0));
+  uint32_t int_st = READ_REG(UART_INT_ST(0));
   while (1) {
-    uint32_t fifo_len = READ_PERI_REG(UART_STATUS(0)) & 0xff;
+    uint32_t fifo_len = READ_REG(UART_STATUS(0)) & 0xff;
     if (fifo_len == 0) {
       break;
     }
     while (fifo_len-- > 0) {
-      uint8_t byte = READ_PERI_REG(UART_FIFO(0)) & 0xff;
+      uint8_t byte = READ_REG(UART_FIFO(0)) & 0xff;
       uart_isr_receive(byte);
     }
   }
-  WRITE_PERI_REG(UART_INT_CLR(0), int_st);
+  WRITE_REG(UART_INT_CLR(0), int_st);
 }
 
 static esp_command_error verify_data_len(esp_command_req_t *command, uint8_t len)
@@ -159,7 +145,7 @@ void cmd_loop() {
     switch(command->op) {
     case ESP_READ_REG:
         if (command->data_len == 4) {
-            resp.value = REG_READ(data_words[0]);
+            resp.value = READ_REG(data_words[0]);
         }
         break;
     case ESP_FLASH_VERIFY_MD5:
@@ -238,6 +224,9 @@ void cmd_loop() {
         break;
     case ESP_FLASH_DATA:
     case ESP_FLASH_DEFLATED_DATA:
+#ifdef ESP32
+    case ESP_FLASH_ENCRYPT_DATA:
+#endif
       /* ACK DATA commands immediately, then process them a few lines down,
          allowing next command to buffer */
       if(is_in_flash_mode()) {
@@ -272,7 +261,7 @@ void cmd_loop() {
       /* params are addr, value, mask (ignored), delay_us (ignored) */
       error = verify_data_len(command, 16);
       if (error == ESP_OK) {
-        REG_WRITE(data_words[0], data_words[1]);
+        WRITE_REG(data_words[0], data_words[1]);
       }
       break;
     case ESP_READ_REG:
@@ -317,6 +306,12 @@ void cmd_loop() {
         /* drop into flashing mode, discard 16 byte payload header */
         handle_flash_data(command->data_buf + 16, command->data_len - 16);
         break;
+#ifdef ESP32
+      case ESP_FLASH_ENCRYPT_DATA:
+        /* write encrypted data */
+        handle_flash_encrypt_data(command->data_buf + 16, command->data_len -16);
+        break;
+#endif
       case ESP_FLASH_DEFLATED_DATA:
         handle_flash_deflated_data(command->data_buf + 16, command->data_len - 16);
         break;
@@ -400,16 +395,18 @@ void stub_main()
   /* All UART reads come via uart_isr */
   ub.reading_buf = ub.buf_a;
   ets_isr_attach(ETS_UART0_INUM, uart_isr, NULL);
-  SET_PERI_REG_MASK(UART_INT_ENA(0), UART_RX_INTS);
+  REG_SET_MASK(UART_INT_ENA(0), UART_RX_INTS);
   ets_isr_unmask(1 << ETS_UART0_INUM);
 
   /* Configure default SPI flash functionality.
      Can be overriden later by esptool.py. */
 #ifdef ESP8266
         SelectSpiFunction();
+
+        spi_flash_attach();
 #else
         uint32_t spiconfig = ets_efuse_get_spiconfig();
-        uint32_t strapping = REG_READ(GPIO_STRAP_REG);
+        uint32_t strapping = READ_REG(GPIO_STRAP_REG);
         /* If GPIO1 (U0TXD) is pulled low and no other boot mode is
            set in efuse, assume HSPI flash mode (same as normal boot)
         */
